package com.googlecode.genericdao.search;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Collection;
import java.util.regex.Pattern;

/**
 * Created by petercipov on 30/06/16.
 */
public class Matcher {

    public enum Operator {
        GREATER,
        GREATER_EQUALS,
        LESS,
        LESS_EQUALS,
        EQUALS
    }

    public static boolean matchesLikeOperation(MatcherContext context, Object matchTo, Object value) {
        if (value == null && matchTo == null) {
            return true;
        }

        if (value == null || matchTo == null) {
            return false;
        }

        if (!(value instanceof String) && !(matchTo instanceof String)) {
            return false;
        }

        Pattern pattern = context.getPattern((String) value);
        if (pattern == null) {
            return false;
        }

        return pattern.matcher((String) matchTo).matches();
    }

    public static boolean matchesOperation(Object value, Object matchTo, Operator operator, boolean ignoreCase) {
        return matchesScalar(value, matchTo, operator, ignoreCase);

    }

    public static boolean matchesIn(Object value, Object matchTo, boolean ignoreCase) {

        if (matchTo == null) {
            return false;
        }

        if (matchTo instanceof Collection) {
            return matchesInCollection(value, (Collection) matchTo, ignoreCase);
        } else if (matchTo instanceof Object[]) {
            return matchesInCollection(value, Arrays.asList((Object[]) matchTo), ignoreCase);
        }

        return false;
    }

    public static boolean matchesInCollection(Object value, Collection matchTo, boolean ignoreCase) {
        for (Object candidate : matchTo) {
            if (matchesScalar(value, candidate, Operator.EQUALS, ignoreCase)) {
                return true;
            }
        }

        return false;
    }

    public static Pattern sqlPatternToPattern(String pattern, boolean ignoreCase) {
        String preprocessed = sqlPatternToRegex(pattern);

        int flag = ignoreCase ? Pattern.CASE_INSENSITIVE : 0;
        return  Pattern.compile(preprocessed, flag);
    }

    public static String sqlPatternToRegex(String pattern) {
        if (pattern == null) {
            throw new NullPointerException("Null pattern.");
        }

        if (pattern.length() == 0) {
            throw new IllegalArgumentException("Empty pattern.");
        }

        StringBuilder buffer = new StringBuilder();

        // convert * into regex syntax
        // e.g. abc*x becomes ^abc.*x$
        // or abc?x becomes ^abc.?x$
        buffer.append("^");
        for (int j = 0; j < pattern.length(); j++) {
            char nextChar = pattern.charAt(j);
            if (nextChar == '%') {
                nextChar = '*';
            }

            if (nextChar == '*' || nextChar == '?') {
                buffer.append('.');
            }
            // escape special chars
            else if (nextChar == '.'
                || nextChar == '/'
                || nextChar == '$'
                || nextChar == '^') {
                buffer.append('\\');
            }

            buffer.append(nextChar);
        }

        buffer.append("$");
        return buffer.toString();
    }

    private static boolean matchesScalar(Object l, Object r, Operator type, boolean ignoreCase) {

        int c = compareScalars(l, r, ignoreCase);

        if (l == null && c != 0) {
            return false;
        }

        // Use the direct values here for speed.

        if (Operator.GREATER.equals(type) && (c < 1)) {
            // The LHS value is equal to or less than
            // the RHS value, but we expect >, so can safely
            // return false.
            return false;
        }

        if (
            (Operator.GREATER_EQUALS.equals(type) || Operator.EQUALS.equals(type))
                && (c < 0)) {

            // The RHS value is less than the LHS value
            // but we expect >=, so can safely return false.
            return false;
        }

        // Check for <
        if (Operator.LESS.equals(type) && (c > -1)) {

            // The RHS value is greater than or equal to
            // the LHS value but we expect <, so can safely return
            // false.

            return false;

        }

        // Check for <= or =
        if ((Operator.LESS_EQUALS.equals(type)
            ||
            Operator.EQUALS.equals(type)
        ) && (c > 0)) {

            // The RHS value is greater than the LHS value
            // but we expect <=, so can safely return false.

            return false;
        }

        return true;
    }

    public static int compareScalars(Object o1, Object o2, boolean ignorecase) {

        if (o1 == null && o2 == null ) {
            return 0;
        }

        if (o1 == null && o2 != null ) {
            return 1;
        }

        if (o1 != null && o2 == null) {
            return -1;
        }

        if ((o1 instanceof BigDecimal) || (o2 instanceof BigDecimal)) {
            return toBigDecimal(o1).compareTo(toBigDecimal(o2));
        }

        if ((o1 instanceof Number) && (o2 instanceof Number)) {
            return toDoubleObject(o1).compareTo(toDoubleObject(o2));
        }

        if ((o1 instanceof Comparable) && (o2 instanceof Comparable)
            && (o1.getClass().isAssignableFrom(o2.getClass())) && !(o1 instanceof String)) {
            return ((Comparable) o1).compareTo(o2);
        }

        // Force a string comparison.
        String s1 = o1.toString();
        String s2 = o2.toString();

        if (ignorecase) {
            s1 = s1.toLowerCase();
            s2 = s2.toLowerCase();
        }

        return s1.compareTo(s2);

    }

    public static Double toDoubleObject(Object o) {
        return new Double(toDouble(o));
    }

    public static double toDouble(Object o) {
        return ((Number) o).doubleValue();
    }

    public static BigDecimal toBigDecimal(Object o) {
        if (o instanceof BigDecimal) {
            return (BigDecimal) o;
        } else {
            return new BigDecimal(o.toString());
        }
    }
}
